/****************************************************************************
**
** Copyright (C) 2023 The Qt Company Ltd.
** Contact: https://www.qt.io/licensing/
**
** This file is part of the Qt Design Viewer of the Qt Toolkit.
**
** Commercial License Usage
** Licensees holding valid commercial Qt licenses may use this file in
** accordance with the commercial license agreement provided with the
** Software or, alternatively, in accordance with the terms contained in
** a written agreement between you and The Qt Company. For licensing terms
** and conditions see https://www.qt.io/terms-conditions. For further
** information use the contact form at https://www.qt.io/contact-us.
**
** GNU General Public License Usage
** Alternatively, this file may be used under the terms of the GNU
** General Public License version 3 as published by the Free Software
** Foundation with exceptions as appearing in the file LICENSE.GPL3-EXCEPT
** included in the packaging of this file. Please review the following
** information to ensure the GNU General Public License requirements will
** be met: https://www.gnu.org/licenses/gpl-3.0.html.
**
****************************************************************************/

#include "projectManager.h"

#include <QBuffer>
#include <QDir>
#include <QDirIterator>
#include <QEventLoop>
#include <QFile>
#include <QFileInfo>
#include <QGuiApplication>
#include <QJsonDocument>
#include <QQuickItem>
#include <QRandomGenerator>
#include <QRegularExpression>
#include <QResource>
#include <QStandardPaths>
#include <QTemporaryDir>
#include <QTemporaryFile>
#include <QtGui/private/qzipreader_p.h>

ProjectManager::ProjectManager(QObject *parent)
    : QObject(parent)
    , m_projectCachePath(QStandardPaths::writableLocation(QStandardPaths::AppDataLocation))
    , m_demoProjectCachePath(m_projectCachePath + "/demoProjects")
{
    qDebug() << "ProjectManager created.";
    qDebug() << "Project cache path: " << m_projectCachePath;
    qDebug() << "Demo project cache path: " << m_demoProjectCachePath;

    if (!QDir(m_projectCachePath).exists()) {
        qDebug() << "Creating project cache path: " << m_projectCachePath;
        QDir().mkpath(m_projectCachePath);
    }

    if (!QDir(m_demoProjectCachePath).exists()) {
        qDebug() << "Creating demo project cache path: " << m_demoProjectCachePath;
        QDir().mkpath(m_demoProjectCachePath);
    }
}

ProjectManager::~ProjectManager()
{
    cleanupResources();
    qDebug() << "ProjectManager destroyed.";
}

void ProjectManager::cleanupResources()
{
    const uchar *data;
    if (m_projectData.size()) {
        qDebug() << "Unregistering the previous data from QRC system. Path: " << m_projectPath
                 << " Size: " << m_projectData.size() << " bytes.";
        data = reinterpret_cast<const uchar *>(m_projectData.data());
        if (!QResource::unregisterResource(data, m_projectPath)) {
            qCritical("Cannot unregister the previous resource data.");
        }
    }
}

QString ProjectManager::unpackProject(const QByteArray &project, bool extractZip)
{
    QTemporaryDir tempDir("qmlprojector");
    QString projectPath = tempDir.path();

    if (extractZip) {
        QDir().mkpath(projectPath);
        QBuffer buffer;
        buffer.setData(project);
        buffer.open(QIODevice::ReadOnly);
        QZipReader reader(&buffer);
        reader.extractAll(projectPath);
    }

    qDebug() << "Initial project location: " << projectPath;

    QDir projectPathDir(projectPath);

    // maybe it was not a zip file so try it as resource binary
    if (projectPathDir.isEmpty()) {
        if (extractZip)
            qDebug("File could not be extracted. Trying to open it as a resource file.");

        cleanupResources();

        m_projectData = project;
        const uchar *data;
        data = reinterpret_cast<const uchar *>(m_projectData.data());
        qDebug() << "Registering resource data. Size: " << m_projectData.size();

        const QString resourcePath{"/" + QString::number(QRandomGenerator::global()->generate())};
        m_projectPath = resourcePath;

        if (!QDir(resourcePath).removeRecursively()) {
            qDebug() << "Could not remove resource path: " << resourcePath;
        }

        if (!QResource::registerResource(data, resourcePath)) {
            qCritical("Can not load the resource data.");
            return "";
        }

        projectPath = ":" + resourcePath;
    }
    return projectPath;
}

QString ProjectManager::findFile(const QString &dir, const QString &filter)
{
    QDirIterator it(dir, {filter}, QDir::Files);
    return it.next();
}

QString ProjectManager::readQmlProjectFile(const QString &qmlProjectFilePath)
{
    QFile file(qmlProjectFilePath);
    if (!file.open(QIODevice::ReadOnly)) {
        qCritical() << "Could not open Qml Project file! " << file.fileName() + ": "
                    << file.errorString();
        return {};
    }

    return QString::fromUtf8(file.readAll());
}

QString ProjectManager::getMainQmlFile(const QString &projectPath,
                                       const QString &qmlProjectFileContent)
{
    const QRegularExpression mainFileRegExp("mainFile:\\s*\"(.*)\"");
    const QRegularExpressionMatch mainFileMatch = mainFileRegExp.match(qmlProjectFileContent);

    if (!mainFileMatch.hasMatch()) {
        qCritical() << "No main file found in " << qmlProjectFileContent;
        return {};
    }

    qDebug() << "Found main file: " << mainFileMatch.captured(1);
    return projectPath + "/" + mainFileMatch.captured(1);
}

QStringList ProjectManager::getImportPaths(const QString &projectPath,
                                           const QString &qmlProjectFileContent)
{
    const QRegularExpression importPathsRegExp("importPaths:\\s*\\[\\s*(.*)\\s*\\]");
    const QRegularExpressionMatch importPathsMatch = importPathsRegExp.match(qmlProjectFileContent);

    if (!importPathsMatch.hasMatch()) {
        qWarning() << "No import paths found in " << qmlProjectFileContent;
        return {};
    }

    QStringList importPaths;
    for (const QString &path : importPathsMatch.captured(1).split(",")) {
        QString cleanedPath = path.trimmed();
        cleanedPath = projectPath + "/" + cleanedPath.mid(1, cleanedPath.length() - 2);
        if (QFileInfo::exists(cleanedPath)) {
            importPaths.append(cleanedPath);
        }
    }

    return importPaths;
}

bool ProjectManager::isQt6Project(const QString &qmlProjectFileContent)
{
    const QRegularExpression qt6ProjectRegExp("qt6Project:\\s*true");
    const QRegularExpressionMatch qt6ProjectMatch = qt6ProjectRegExp.match(qmlProjectFileContent);
    return qt6ProjectMatch.hasMatch();
}

bool ProjectManager::runProject(const QString &projectPath)
{
    qDebug() << "Project location: " << projectPath;

    const QString qmlProjectFile = findFile(projectPath, "*.qmlproject");
    if (qmlProjectFile.isEmpty()) {
        qCritical() << "No \"*.qmlproject\" found in \"" << projectPath << "\".";
        return false;
    }

    const QString qmlProjectFileContent = readQmlProjectFile(qmlProjectFile);
    if (qmlProjectFileContent.isEmpty())
        return false;

    const QString mainQmlFilePath = getMainQmlFile(projectPath, qmlProjectFileContent);
    if (mainQmlFilePath.isEmpty())
        return false;

    const QString qtquickcontrols2File = findFile(projectPath, "qtquickcontrols2.conf");
    if (!qtquickcontrols2File.isEmpty()) {
        qputenv("QT_QUICK_CONTROLS_CONF", qtquickcontrols2File.toLatin1());
    }

    const QStringList importPaths = getImportPaths(projectPath, qmlProjectFileContent);
    if (isQt6Project(qmlProjectFileContent)) {
        qWarning() << "This is not a Qt6 project.\nQt5 projects might work, but they are not "
                      "officially supported.";
    }

    qDebug() << "Project is parsed successfully.";
    qDebug() << "Qml Project File: " << qmlProjectFile;
    qDebug() << "Main Qml File: " << mainQmlFilePath;
    qDebug() << "Import Paths: " << importPaths;
    qDebug() << "Qt Quick Controls 2 File: " << qtquickcontrols2File;

    QUrl mainQmlUrl = QUrl::fromUserInput(mainQmlFilePath);
    QFile file(mainQmlUrl.path());
    if (!file.open(QIODevice::ReadOnly)) {
        qCritical() << "Could not open mainQmlfile for reading! " << file.fileName() << ": "
                    << file.errorString();
        return false;
    }

    qDebug() << "Initializing the qmlEngine";
    m_qmlEngine.reset(new QQmlEngine);
    m_qmlEngine->clearComponentCache();

    qDebug("Adding import paths");
    for (const QString &importPath : importPaths) {
        qDebug() << "-- Import path: " << importPath;
        m_qmlEngine->addImportPath(importPath);
    }

    QObject::connect(m_qmlEngine.data(),
                     &QQmlEngine::warnings,
                     this,
                     [&](const QList<QQmlError> &warnings) {
                         for (const auto &warning : warnings) {
                             qWarning() << warning.toString();
                         }
                     });

    qDebug() << "Loading mainQmlUrl: " << mainQmlUrl.toString();
    m_qmlComponent.reset(new QQmlComponent(m_qmlEngine.data()));

    m_qmlComponent->loadUrl(mainQmlUrl);

    qDebug() << "Waiting for qmlComponent to load";
    while (m_qmlComponent->isLoading())
        QCoreApplication::processEvents();

    qDebug() << "Checking if m_qmlComponent is ready";
    if (!m_qmlComponent->isReady()) {
        qCritical() << "m_qmlComponent is not ready. Reason:" << m_qmlComponent->errorString();
        return false;
    }
    qDebug() << "Creating top level object";
    QObject *topLevel = m_qmlComponent->create();
    if (!topLevel && m_qmlComponent->isError()) {
        qCritical() << "Error while creating Qml m_qmlComponent:" << m_qmlComponent->errorString();
        return false;
    }

    qDebug() << "Setting up the quickWindow";
    m_quickWindow.reset(qobject_cast<QQuickWindow *>(topLevel));
    if (m_quickWindow) {
        qDebug() << "Running with incubator controller";
        m_qmlEngine->setIncubationController(m_quickWindow->incubationController());
    } else {
        qWarning() << "Top level object is not a QQuickWindow. Trying QQuickView...";

        QQuickItem *contentItem = qobject_cast<QQuickItem *>(topLevel);
        if (!contentItem) {
            qCritical() << "Top level object cannot be casted to QQuickItem. Aborting.";
            return false;
        }

        qDebug() << "Initializing QQuickView";
        QQuickView *view = new QQuickView(m_qmlEngine.data(), nullptr);
        m_quickWindow.reset(view);
        view->setContent(mainQmlUrl, m_qmlComponent.data(), contentItem);
        view->setResizeMode(QQuickView::SizeViewToRootObject);
        m_quickWindow->setBaseSize(QSize(contentItem->width(), contentItem->height()));
    }
    return true;
}

bool ProjectManager::cacheProject(const QByteArray &projectData, const QJsonObject &projectInfo)
{
    // sample project info
    /*
     {
        "appName": "Testmcu",
        "id": "0sadf8fa9s8df67s8998690a7sdf",
        "owner": "",
        "passwordHash": "",
        "qdsIsEnterprise": true,
        "qdsVersion": "Qt Design Studio 4.2.0",
        "ttlDays": 31,
        "uploadTime": "2023-10-27T13:57:22",
        "userHash": "12038740912873462987"
      },
    */

    const QString projectId = projectInfo.value("id").toString();

    qDebug() << "Caching project " << projectId << " with last modified "
             << projectInfo.value("uploadTime");
    const QString cachePath = m_projectCachePath + "/" + projectId;

    // remove old cache
    if (QDir(cachePath).exists()) {
        qDebug() << "Removing old cache for project " << projectId;
        if (!QDir(cachePath).removeRecursively()) {
            qCritical() << "Could not remove old cache for project " << projectId;
            return false;
        }
    }

    QDir().mkpath(cachePath);
    const QString tempProjectPath = unpackProject(projectData);

    // copy all files from tempProjectPath to cachePath
    QDirIterator it(tempProjectPath, QDirIterator::Subdirectories);
    while (it.hasNext()) {
        it.next();
        const QString filePath = it.filePath();
        const QString relativeFilePath = filePath.mid(tempProjectPath.length());
        const QString newFilePath = cachePath + relativeFilePath;

        if (QFileInfo(filePath).isDir()) {
            QDir().mkpath(newFilePath);
        } else {
            QFile::copy(filePath, newFilePath);
        }
    }

    // write project info to cache
    QFile lastModifiedFile(cachePath + "/projectInfo.json");
    if (!lastModifiedFile.open(QIODevice::WriteOnly)) {
        qCritical() << "Could not open projectInfo file for writing";
        return false;
    }

    lastModifiedFile.write(QJsonDocument(projectInfo).toJson());
    return true;
}

bool ProjectManager::isProjectCached(const QJsonObject &projectInfo)
{
    const QString projectId = projectInfo.value("id").toString();
    const QString lastModified = projectInfo.value("uploadTime").toString();

    qDebug() << "Checking if project " << projectId << " is cached";
    const QString cachePath = m_projectCachePath + "/" + projectId;
    if (!QDir(cachePath).exists()) {
        qDebug() << "Project " << projectId << " is not cached";
        return false;
    }

    qDebug() << "Project " << projectId << " is cached. Checking if it is up to date";
    QFile lastModifiedFile(cachePath + "/projectInfo.json");
    if (!lastModifiedFile.open(QIODevice::ReadOnly)) {
        qCritical() << "Could not open projectInfo file for reading";
        return false;
    }

    const QJsonObject cachedProjectInfo
        = QJsonDocument::fromJson(lastModifiedFile.readAll()).object();
    const QString cachedLastModified = cachedProjectInfo.value("uploadTime").toString();

    qDebug() << "Project " << projectId << " last modified: " << lastModified;
    qDebug() << "Cached project " << projectId << " last modified: " << cachedLastModified;

    if (lastModified != cachedLastModified) {
        qDebug() << "Project " << projectId << " is not up to date";
        return false;
    }

    qDebug() << "Project " << projectId << " is up to date";
    return true;
}

bool ProjectManager::runCachedProject(const QJsonObject &projectInfo)
{
    const QString projectId = projectInfo.value("id").toString();
    qDebug() << "Running cached project " << projectId;

    const QString projectPath = m_projectCachePath + "/" + projectId;
    return runProject(projectPath);
}

bool ProjectManager::cacheDemoProject(const QByteArray &projectData, const QJsonObject &projectInfo)
{
    // sample project info
    //[{"lastUpdate":1701947766739.9812,"name":"ClusterTutorial.qmlrc"}]
    const QString projectName = projectInfo.value("name").toString().remove(".qmlrc");
    const QString demoProjectPath = m_demoProjectCachePath + "/" + projectName;
    qDebug() << "Caching demo project " << projectName << " to " << demoProjectPath;

    // remove old cache
    if (QDir(demoProjectPath).exists()) {
        qDebug() << "Removing old cache for demo project " << projectName;
        if (!QDir(demoProjectPath).removeRecursively()) {
            qCritical() << "Could not remove old cache for demo project " << projectName;
            return false;
        }
    }

    // create new cache
    if (!QDir().mkpath(demoProjectPath)) {
        qCritical() << "Could not create new cache for demo project " << projectName;
        return false;
    }

    const QString tempProjectPath = unpackProject(projectData);

    // copy all files from tempProjectPath to demoProjectPath
    QDirIterator it(tempProjectPath, QDirIterator::Subdirectories);
    while (it.hasNext()) {
        it.next();
        const QString filePath = it.filePath();
        const QString relativeFilePath = filePath.mid(tempProjectPath.length());
        const QString newFilePath = demoProjectPath + relativeFilePath;

        if (QFileInfo(filePath).isDir()) {
            QDir().mkpath(newFilePath);
        } else if (!QFile::copy(filePath, newFilePath)) {
            qCritical() << "Could not copy " << filePath << " to " << newFilePath;
            return false;
        }
    }

    // write project info to cache
    QFile demoProjectInfoFile(demoProjectPath + "/projectInfo.json");
    if (!demoProjectInfoFile.open(QIODevice::WriteOnly)) {
        qCritical() << "Could not open demo project info file for writing";
        return false;
    }

    demoProjectInfoFile.write(QJsonDocument(projectInfo).toJson());
    return true;
}

void ProjectManager::clearDemoCaches()
{
    qDebug() << "Clearing demo caches";
    QDir(m_demoProjectCachePath).removeRecursively();
}

bool ProjectManager::isDemoProjectCached(const QJsonObject &projectInfo)
{
    // sample project info
    //[{"lastUpdate":1701947766739.9812,"name":"ClusterTutorial.qmlrc"}]
    const QString projectName = projectInfo.value("name").toString().remove(".qmlrc");
    const QString demoProjectPath = m_demoProjectCachePath + "/" + projectName;
    qDebug() << "Checking if demo project " << projectName << " is cached";

    if (!QDir(demoProjectPath).exists()) {
        qDebug() << "Demo project " << projectName << " is not cached";
        return false;
    }

    qDebug() << "Demo project " << projectName << " is cached. Checking the last update time...";
    const QString demoProjectInfoPath = demoProjectPath + "/projectInfo.json";
    QFile demoProjectInfoFile(demoProjectInfoPath);
    if (!demoProjectInfoFile.open(QIODevice::ReadOnly)) {
        qCritical() << "Could not open demo project info file for reading";
        return false;
    }

    const QJsonObject cachedProjectInfo
        = QJsonDocument::fromJson(demoProjectInfoFile.readAll()).object();

    const QJsonValue cachedLastModified = cachedProjectInfo.value("lastUpdate");
    const QJsonValue cloudLastModified = projectInfo.value("lastUpdate");
    qDebug() << "Demo project" << projectName << "cloud timestamp: " << cloudLastModified
             << "local timestamp:" << cachedLastModified;

    if (cloudLastModified != cachedLastModified) {
        qDebug() << "Demo project " << projectName << " is not up to date";
        return false;
    }

    qDebug() << "Demo project " << projectName << " is up to date";
    return true;
}

bool ProjectManager::runDemoProject(const QString &projectName)
{
    const QString demoProjectPath = m_demoProjectCachePath + "/" + projectName;
    qDebug() << "Running demo project " << projectName << " from " << demoProjectPath;
    return runProject(demoProjectPath);
}

void ProjectManager::orientateWindow(Qt::ScreenOrientation orientation)
{
    QQuickItem *contentItem = m_quickWindow->contentItem();
    QQuickItem *childItem{contentItem->childItems().at(0)};

    const QRect screenGeometry = QGuiApplication::primaryScreen()->geometry();
    const QString orientationString = (orientation == Qt::LandscapeOrientation ? "landscape"
                                                                               : "portrait");
    qDebug() << "Adapting orientation to " << orientationString << " mode"
             << ". Initial sizing:" << Qt::endl
             << "-- Screen size: " << screenGeometry.height() << " x " << screenGeometry.width()
             << Qt::endl
             << "-- Quick window size: " << m_quickWindow->height() << " x "
             << m_quickWindow->width() << Qt::endl
             << "-- Child size: " << childItem->height() << " x " << childItem->width() << Qt::endl
             << "-- Child pos: " << childItem->x() << ", " << childItem->y() << Qt::endl
             << "-- Child scale: " << childItem->scale();

    const QSizeF newContentSize = childItem->size().scaled(screenGeometry.size().toSizeF(),
                                                           Qt::AspectRatioMode::KeepAspectRatio);

    const qreal newScale = newContentSize.height() / childItem->size().height();
    const qreal newX = (childItem->width() - screenGeometry.width()) / -2.0f;
    const qreal newY = (childItem->height() - screenGeometry.height()) / -2.0f;

    childItem->setScale(newScale);
    childItem->setPosition(QPointF(newX, newY));

    qDebug() << "Calculated props: height-width-scale-posX-posY" << Qt::endl
             << "--" << newContentSize.height() << "-" << newContentSize.width() << "-" << newScale
             << "-" << newX << "-" << newY << Qt::endl
             << "Final Sizing: height-width-scale-posX-posY" << Qt::endl
             << "--" << childItem->height() << "-" << childItem->width() << "-"
             << childItem->scale() << "-" << childItem->x() << "-" << childItem->y();
}

void ProjectManager::showAppWindow()
{
    qDebug("Initializing and showing the QML app window");

    QScreen *screen = QGuiApplication::primaryScreen();
    connect(screen, &QScreen::orientationChanged, this, &ProjectManager::orientateWindow);
    orientateWindow(screen->orientation());

    connect(m_quickWindow.data(), &QQuickWindow::closing, this, [this, screen]() {
        qDebug() << "QML app window is closing";
        disconnect(screen, &QScreen::orientationChanged, this, &ProjectManager::orientateWindow);
        // this signal is connected to the lambda in the backend
        // which will reset the project manager
        emit closingProject();
    });

    m_quickWindow->setFlags(Qt::Window | Qt::WindowStaysOnTopHint);
    m_quickWindow->showMaximized();
}

void ProjectManager::hideAppWindow()
{
    qDebug("Hiding the QML app window");
    m_quickWindow->hide();
}
